;;; finder-inf.el --- keyword-to-package mapping
;; Keywords: help
;;; Commentary:
;; Don't edit this file.  It's generated by finder.el

;;; Code:

(setq finder-package-info '(
    ("abbrev.el"
        "abbrev mode commands for Emacs"
        (abbrev))
    ("ada.el"
        "Ada editing support package in GNUlisp.  v1.0"
        (languages))
    ("add-log.el"
        "change log maintenance commands for Emacs"
        (maint))
    ("advice.el"
        "advice mechanism for Emacs Lisp functions"
        (advice  function hooks))
    ("allout.el"
        nil
        nil)
    ("ange-ftp.el"
        "transparent FTP support for GNU Emacs"
        (comm))
    ("appt.el"
        "appointment notification functions."
        (calendar))
    ("apropos.el"
        "faster apropos commands."
        (help))
    ("array.el"
        "array editing commands for Gnu Emacs"
        (extensions))
    ("asm-mode.el"
        "mode for editing assembler code"
        (tools  languages))
    ("assoc.el"
        "insert/delete/sort functions on association lists"
        (extensions))
    ("autoinsert.el"
        "automatic mode-dependent insertion of text into new files"
        nil)
    ("autoload.el"
        "maintain autoloads in loaddefs.el."
        (maint))
    ("awk-mode.el"
        "AWK code editing commands for Emacs"
        (unix  languages))
    ("backquote.el"
        "backquoting for Emacs Lisp macros"
        (extensions))
    ("bib-mode.el"
        "bib-mode, major mode for editing bib files."
        (bib))
    ("bibtex.el"
        "BibTeX mode for GNU Emacs"
        (tex  bib))
    ("blackbox.el"
        "blackbox game in Emacs Lisp"
        (games))
    ("bookmark.el"
        "set bookmarks, jump to them later."
        (bookmarks  placeholders))
    ("buff-menu.el"
        "buffer menu main function and support functions."
        nil)
    ("byte-opt.el"
        "the optimization passes of the emacs-lisp byte compiler."
        (internal))
    ("byte-run.el"
        "byte-compiler support for inlining"
        (internal))
    ("bytecomp.el"
        "compilation of Lisp code into byte code."
        (internal))
    ("c-mode.el"
        "C code editing commands for Emacs"
        (c))
    ("cal-dst.el"
        "calendar functions for daylight savings rules."
        (calendar))
    ("cal-french.el"
        "calendar functions for the French Revolutionary calendar."
        (calendar))
    ("cal-mayan.el"
        "calendar functions for the Mayan calendars."
        (calendar))
    ("calendar.el"
        "Calendar functions."
        (calendar))
    ("case-table.el"
        "code to extend the character set and support case tables."
        (i18n))
    ("chistory.el"
        "list command history"
        nil)
    ("cl-compat.el"
        nil
        (extensions))
    ("cl-extra.el"
        nil
        (extensions))
    ("cl-indent.el"
        "enhanced lisp-indent mode"
        (lisp  tools))
    ("cl-macs.el"
        nil
        (extensions))
    ("cl-seq.el"
        nil
        (extensions))
    ("cl.el"
        nil
        (extensions))
    ("cmacexp.el"
        "expand C macros in a region"
        (c))
    ("cmuscheme.el"
        "Scheme process in a buffer. Adapted from tea.el."
        (processes  lisp))
    ("comint.el"
        "general command interpreter in a window stuff"
        (processes))
    ("compare-w.el"
        "compare text between windows for Emacs."
        nil)
    ("compile.el"
        "run compiler as inferior of Emacs, parse error messages."
        (tools  processes))
    ("complete.el"
        nil
        nil)
    ("completion.el"
        "dynamic word-completion code"
        (abbrev))
    ("cookie1.el"
        "retrieve random phrases from fortune cookie files"
        (games))
    ("cplus-md.el"
        "C++ code editing commands for Emacs"
        (c))
    ("cplus-md1.el"
        "major mode for editing C++ (and C) code"
        (c++ c editing major-mode))
    ("cust-print.el"
        "handles print-level and print-circle."
        (extensions))
    ("dabbrev.el"
        "dynamic abbreviation package for GNU Emacs."
        (abbrev))
    ("debug.el"
        "debuggers and related commands for Emacs"
        (lisp  tools  maint))
    ("default.el"
        nil
        nil)
    ("delsel.el"
        "delete selection if you insert"
        nil)
    ("desktop.el"
        "save partial status of Emacs when killed"
        nil)
    ("diary-ins.el"
        "calendar functions for adding diary entries."
        (diary  calendar))
    ("diary.el"
        "diary functions."
        (calendar))
    ("diff.el"
        "Run `diff' in compilation-mode."
        (unix  tools))
    ("dired-aux.el"
        "all of dired except what people usually use"
        nil)
    ("dired.el"
        "directory-browsing commands"
        nil)
    ("disass.el"
        "disassembler for compiled Emacs Lisp code"
        (internal))
    ("disp-table.el"
        "functions for dealing with char tables."
        (i18n))
    ("dissociate.el"
        "scramble text amusingly for Emacs."
        (games))
    ("doctor.el"
        "psychological help for frustrated users."
        (games))
    ("dunnet.el"
        "Text adventure for Emacs"
        (games))
    ("ebuff-menu.el"
        "electric-buffer-list mode"
        nil)
    ("echistory.el"
        "Electric Command History Mode"
        nil)
    ("edebug.el"
        "a source-level debugger for emacs lisp"
        (lisp  tools  maint))
    ("edmacro.el"
        "keyboard macro editor"
        (abbrev))
    ("edt.el"
        "EDT emulation in Emacs"
        (emulations))
    ("ehelp.el"
        "bindings for electric-help mode"
        (help  extensions))
    ("electric.el"
        "window maker and Command loop for `electric' modes."
        (extensions))
    ("emacsbug.el"
        "command to report Emacs bugs to appropriate mailing list."
        (maint))
    ("emerge.el"
        "merge diffs under Emacs control"
        (unix  tools))
    ("env.el"
        "functions to manipulate environment variables."
        (processes  unix))
    ("etags.el"
        "etags facility for Emacs"
        (tools))
    ("faces.el"
        "Lisp interface to the c \"face\" structure"
        nil)
    ("files.el"
        "file input and output commands for Emacs"
        nil)
    ("fill.el"
        "fill commands for Emacs"
        (wp))
    ("find-dired.el"
        "run a `find' command and dired the output"
        (unix))
    ("find-gc.el"
        "detect functions that call the garbage collector"
        nil)
    ("finder-inf.el"
        "keyword-to-package mapping"
        (help))
    ("finder.el"
        "topic & keyword-based code finder"
        (help))
    ("float-sup.el"
        "detect absence of floating-point support in Emacs runtime"
        nil)
    ("float.el"
        "floating point arithmetic package."
        (extensions))
    ("flow-ctrl.el"
        "help for lusers on cu(1) or ttys with wired-in ^S/^Q flow control"
        (hardware))
    ("font-lock.el"
        nil
        (languages  faces))
    ("forms-d2.el"
        nil
        nil)
    ("forms-pass.el"
        nil
        nil)
    ("forms.el"
        nil
        nil)
    ("fortran.el"
        "Fortran mode for GNU Emacs"
        (languages))
    ("frame.el"
        "multi-frame management independent of window systems."
        (internal))
    ("gnus.el"
        nil
        (news))
    ("gnusmail.el"
        "mail reply commands for GNUS newsreader"
        (news))
    ("gnusmisc.el"
        "miscellaneous commands for GNUS newsreader"
        (news))
    ("gnuspost.el"
        "post news commands for GNUS newsreader"
        (news))
    ("gomoku.el"
        "Gomoku game between you and Emacs"
        (games))
    ("gosmacs.el"
        "rebindings to imitate Gosmacs."
        (emulations))
    ("grow-vers.el"
        "increment Emacs version number"
        (internal))
    ("gud.el"
        "Grand Unified Debugger mode for gdb, sdb, dbx, or xdb"
        (unix  tools))
    ("hanoi.el"
        "towers of hanoi in GNUmacs"
        (games))
    ("help-macro.el"
        "Makes command line help such as help-for-help"
        nil)
    ("help.el"
        "help commands for Emacs"
        (help  internal))
    ("helper.el"
        "utility help package supporting help in electric modes"
        (help))
    ("hexl.el"
        "edit a file in a hex dump format using the hexl filter."
        (non-text))
    ("hideif.el"
        "hides selected code within ifdef."
        (c))
    ("hilit19.el"
        nil
        (faces))
    ("hippie-exp.el"
        "expand text trying various ways to find its expansion."
        (extensions))
    ("holidays.el"
        "holiday functions for the calendar package"
        (holidays  calendar))
    ("icon.el"
        "mode for editing Icon code"
        (languages))
    ("inc-vers.el"
        "load this to increment the recorded Emacs version number."
        (internal))
    ("indent.el"
        "indentation commands for Emacs"
        nil)
    ("inf-lisp.el"
        "an inferior-lisp mode"
        (processes  lisp))
    ("info.el"
        "info package for Emacs."
        (help))
    ("informat.el"
        "info support functions package for Emacs"
        (help))
    ("isearch.el"
        "incremental search minor mode."
        nil)
    ("iso-ascii.el"
        "set up char tables for ISO 8859/1 on ASCII terminals."
        (i18n))
    ("iso-insert.el"
        "insert functions for ISO 8859/1."
        (i18n))
    ("iso-swed.el"
        "set up char tables for ISO 8859/1 for Swedish/Finnish ttys"
        (i18n))
    ("iso-syntax.el"
        "set up case-conversion and syntax tables for ISO 8859/1"
        (i18n))
    ("ispell.el"
        "this is the GNU EMACS interface to GNU ISPELL version 4."
        (wp))
    ("kermit.el"
        "additions to shell mode for use with kermit, etc."
        (comm))
    ("ledit.el"
        "Emacs side of ledit interface"
        nil)
    ("levents.el"
        "emulate the Lucid event data type and associated functions."
        nil)
    ("life.el"
        "John Horton Conway's `Life' game for GNU Emacs"
        (games))
    ("lisp-mnt.el"
        "minor mode for Emacs Lisp maintainers"
        (docs))
    ("lisp-mode.el"
        "Lisp mode, and its idiosyncratic commands."
        (lisp  languages))
    ("lisp.el"
        "Lisp editing commands for Emacs"
        (lisp  languages))
    ("lmenu.el"
        "emulate Lucid's menubar support"
        (emulations))
    ("loaddefs.el"
        "define standard autoloads and keys of other files"
        (internal))
    ("loadhist.el"
        "lisp functions for working with feature groups"
        (internal))
    ("loadup.el"
        "load up standardly loaded Lisp files for Emacs."
        (internal))
    ("lpr.el"
        "print Emacs buffer on line printer."
        (unix))
    ("ls-lisp.el"
        "emulate insert-directory completely in Emacs Lisp"
        (unix))
    ("lselect.el"
        "Lucid interface to X Selections"
        (emulations))
    ("lucid.el"
        "Emulate some Lucid Emacs functions."
        nil)
    ("lunar.el"
        "calendar functions for phases of the moon."
        (calendar))
    ("macros.el"
        "non-primitive commands for keyboard macros."
        (abbrev))
    ("mail-extr.el"
        "extract full name and address from RFC 822 mail header."
        (mail))
    ("mail-utils.el"
        "utility functions used both by rmail and rnews"
        (mail  news))
    ("mailabbrev.el"
        "abbrev-expansion of mail aliases."
        (mail))
    ("mailalias.el"
        "expand mailing address aliases defined in ~/.mailrc."
        (mail))
    ("mailpost.el"
        "RMAIL coupler to /usr/uci/post mailer"
        (mail))
    ("makefile.el"
        "makefile editing commands for Emacs"
        (unix  tools))
    ("makeinfo.el"
        nil
        nil)
    ("makesum.el"
        "generate key binding summary for Emacs"
        (help))
    ("man.el"
        "browse UNIX manual pages"
        nil)
    ("map-ynp.el"
        "General-purpose boolean question-asker."
        (lisp  extensions))
    ("medit.el"
        "front-end to the MEDIT package for editing MDL"
        (languages))
    ("meese.el"
        "protect the impressionable young minds of America"
        (games))
    ("menu-bar.el"
        "define a default menu bar."
        (internals))
    ("metamail.el"
        "Metamail interface for GNU Emacs"
        (mail  news  mime  multimedia))
    ("mh-e.el"
        "GNU Emacs interface to the MH mail system"
        (mail))
    ("mhspool.el"
        "MH folder access using NNTP for GNU Emacs"
        (mail  news))
    ("mim-mode.el"
        "Mim (MDL in MDL) mode."
        (languages))
    ("mim-syntax.el"
        "syntax checker for Mim (MDL)."
        (languages))
    ("misc.el"
        "basic editing commands for Emacs"
        nil)
    ("mlconvert.el"
        "convert buffer of Mocklisp code to real lisp."
        (emulations))
    ("mlsupport.el"
        "run-time support for mocklisp code."
        (extensions))
    ("modula2.el"
        "Modula-2 editing support package"
        (languages))
    ("mouse.el"
        "window system-independent mouse support."
        (hardware))
    ("mpuz.el"
        "multiplication puzzle for GNU Emacs"
        (games))
    ("netunam.el"
        "HP-UX RFA Commands"
        (comm))
    ("nnspool.el"
        "spool access using NNTP for GNU Emacs"
        (news))
    ("nntp.el"
        "NNTP (RFC977) Interface for GNU Emacs"
        (news))
    ("novice.el"
        "handling of disabled commands (\"novice mode\") for Emacs."
        (internal  help))
    ("nroff-mode.el"
        "GNU Emacs major mode for editing nroff source"
        (wp))
    ("options.el"
        "edit Options command for Emacs."
        nil)
    ("outline.el"
        "outline mode commands for Emacs"
        nil)
    ("page-ext.el"
        "page handling commands"
        nil)
    ("page.el"
        "page motion commands for emacs."
        nil)
    ("paragraphs.el"
        "paragraph and sentence parsing."
        (wp))
    ("paren.el"
        "highlight matching paren."
        (languages  faces))
    ("paths.el"
        "define pathnames for use by various Emacs commands."
        (internal))
    ("perl-mode.el"
        "Perl code editing commands for GNU Emacs"
        (languages))
    ("picture.el"
        "\"Picture mode\" -- editing using quarter-plane screen model."
        nil)
    ("profile.el"
        "generate run time measurements of Emacs Lisp functions"
        (lisp  tools))
    ("prolog.el"
        "major mode for editing and running Prolog under Emacs"
        (languages))
    ("rect.el"
        "rectangle functions for GNU Emacs."
        (internal))
    ("refbib.el"
        "convert refer-style references to ones usable by Latex bib"
        (bib  tex))
    ("refer.el"
        "look up references in bibliography files."
        (bib))
    ("register.el"
        "register commands for Emacs."
        (internal))
    ("replace.el"
        "replace commands for Emacs."
        nil)
    ("reporter.el"
        "customizable bug reporting of lisp programs"
        (bug reports lisp))
    ("reposition.el"
        "center a Lisp function or comment on the screen"
        nil)
    ("resume.el"
        "process command line args from within a suspended Emacs job"
        (processes))
    ("rfc822.el"
        "hairy rfc822 parser for mail and news and suchlike"
        (mail))
    ("ring.el"
        "handle rings of items"
        (extensions))
    ("rlogin.el"
        "remote login interface"
        (unix  comm))
    ("rmail.el"
        "main code of \"RMAIL\" mail reader for Emacs."
        (mail))
    ("rmailedit.el"
        "\"RMAIL edit mode\"  Edit the current message."
        (mail))
    ("rmailkwd.el"
        "part of the \"RMAIL\" mail reader for Emacs."
        (mail))
    ("rmailmsc.el"
        "miscellaneous support functions for the RMAIL mail reader"
        (mail))
    ("rmailout.el"
        "\"RMAIL\" mail reader for Emacs: output message to a file."
        (mail))
    ("rmailsort.el"
        "Rmail: sort messages."
        (mail))
    ("rmailsum.el"
        "make summary buffers for the mail reader"
        (mail))
    ("rnews.el"
        "USENET news reader for gnu emacs"
        (news))
    ("rnewspost.el"
        "USENET news poster/mailer for GNU Emacs"
        (mail  news))
    ("rot13.el"
        "display a buffer in rot13."
        nil)
    ("saveplace.el"
        "automatically save place in files."
        (bookmarks  placeholders))
    ("sc-elec.el"
        nil
        nil)
    ("sc.el"
        nil
        nil)
    ("scheme.el"
        "Scheme mode, and its idiosyncratic commands."
        (languages  lisp))
    ("scribe.el"
        "scribe mode, and its idiosyncratic commands."
        (wp))
    ("scroll-bar.el"
        "window system-independent scroll bar support."
        (hardware))
    ("select.el"
        "lisp portion of standard selection support."
        (internal))
    ("sendmail.el"
        "mail sending commands for Emacs."
        (mail))
    ("server.el"
        "Lisp code for GNU Emacs running as server process."
        (processes))
    ("sgml-mode.el"
        "SGML-editing mode"
        (wp))
    ("shell.el"
        "specialized comint.el for running the shell."
        (processes))
    ("simple.el"
        "basic editing commands for Emacs"
        nil)
    ("simula.el"
        "SIMULA 87 code editing commands for Emacs"
        (languages))
    ("site-load.el"
        nil
        nil)
    ("solar.el"
        "calendar functions for solar events."
        (calendar))
    ("sort.el"
        "commands to sort text in an Emacs buffer."
        (unix))
    ("spell.el"
        "spelling correction interface for Emacs."
        (wp  unix))
    ("spook.el"
        "spook phrase utility for overloading the NSA line eater"
        (games))
    ("startup.el"
        "process Emacs shell arguments"
        (internal))
    ("studly.el"
        "StudlyCaps (tm)(r)(c)(xxx)"
        (games))
    ("subr.el"
        "basic lisp subroutines for Emacs"
        nil)
    ("swedish.el"
        "miscellaneous functions for dealing with Swedish."
        (i18n))
    ("tabify.el"
        "tab conversion commands for Emacs"
        nil)
    ("tar-mode.el"
        "simple editing of tar files from GNU emacs"
        (unix))
    ("tcp.el"
        nil
        nil)
    ("telnet.el"
        "run a telnet session from within an Emacs buffer"
        nil)
    ("term-nasty.el"
        "Damned Things from terminfo.el"
        nil)
    ("terminal.el"
        "terminal emulator for GNU Emacs."
        (comm  terminals))
    ("tex-mode.el"
        "TeX, LaTeX, and SliTeX mode commands."
        (tex))
    ("texinfmt.el"
        nil
        nil)
    ("texinfo.el"
        nil
        nil)
    ("texnfo-upd.el"
        nil
        (maint  tex  docs))
    ("text-mode.el"
        "text mode, and its idiosyncratic commands."
        nil)
    ("time-stamp.el"
        "Maintain last change time stamps in files edited by Emacs"
        (tools))
    ("time.el"
        "display time and load in mode line of Emacs."
        nil)
    ("timer.el"
        "run a function with args at some time in future"
        nil)
    ("timezone.el"
        nil
        nil)
    ("tpu-doc.el"
        "Documentation for TPU-edt"
        (tpu-edt))
    ("tpu-edt.el"
        "Emacs emulating TPU emulating EDT"
        (edt emulations))
    ("tpu-extras.el"
        "Scroll margins and free cursor mode for TPU-edt"
        (tpu-edt))
    ("tpu-mapper.el"
        nil
        (tpu-edt))
    ("tq.el"
        "utility to maintain a transaction queue"
        (extensions))
    ("trace.el"
        "tracing facility for Emacs Lisp functions"
        (tracing  debugging))
    ("two-column.el"
        "minor mode for editing of two-column text"
        nil)
    ("uncompress.el"
        "auto-decompression hook for visiting .Z files"
        nil)
    ("underline.el"
        "insert/remove underlining (done by overstriking) in Emacs."
        (wp))
    ("undigest.el"
        "digest-cracking support for the RMAIL mail reader"
        (mail))
    ("unrmail.el"
        "convert Rmail files to mailbox files."
        (mail))
    ("unused.el"
        "editing commands in GNU Emacs that turned out not to be used."
        (emulations))
    ("upd-copyr.el"
        "update the copyright notice in a GNU Emacs Lisp file"
        (maint))
    ("userlock.el"
        "handle file access contention between multiple users"
        (internal))
    ("vc-hooks.el"
        "resident support for version-control"
        nil)
    ("vc.el"
        "drive a version-control system from within Emacs"
        nil)
    ("version.el"
        "record version number of Emacs."
        (internal))
    ("vi.el"
        "major mode for emulating \"vi\" editor under GNU Emacs."
        (emulations))
    ("view.el"
        "peruse file or buffer without editing."
        nil)
    ("vip.el"
        "a VI Package for GNU Emacs"
        (emulations))
    ("vms-patch.el"
        "override parts of files.el for VMS."
        (vms))
    ("vms-pmail.el"
        "use Emacs as the editor within VMS mail."
        (vms))
    ("vmsproc.el"
        "run asynchronous VMS subprocesses under Emacs"
        (vms))
    ("vmsx.el"
        "run asynchronous VMS subprocesses under Emacs"
        (vms))
    ("vt-control.el"
        "Common VTxxx control functions"
        (vt100))
    ("vt100-led.el"
        "functions for LED control on VT-100 terminals & clones."
        (hardware))
    ("window.el"
        "GNU Emacs window commands aside from those written in C."
        nil)
    ("ws-mode.el"
        "WordStar emulation mode for GNU Emacs"
        (emulations))
    ("x-menu.el"
        "menu support for X "
        nil)
    ("xscheme.el"
        "run Scheme under Emacs"
        (languages lisp))
    ("yow.el"
        "quote random zippyisms"
        (games))
))

(provide 'finder-inf)

;;; finder-inf.el ends here
